# SPDX-FileCopyrightText: 2025 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import asyncio
import unittest
from datetime import datetime, timedelta
from unittest.mock import AsyncMock, MagicMock, patch
from uuid import UUID, uuid4

from rich.console import Console
from rich.progress import Progress

from greenbone.scap.cli import DEFAULT_VERBOSITY, CLIError
from greenbone.scap.cpe_match.cli.processor import CpeMatchProcessor
from greenbone.scap.cpe_match.db.manager import CPEMatchStringDatabaseManager
from greenbone.scap.cpe_match.db.models import (
    CPEMatchDatabaseModel,
    CPEMatchStringDatabaseModel,
)
from greenbone.scap.cpe_match.producer.db import CpeMatchDatabaseProducer
from tests.cpe_match.producer.mock_worker import CpeMatchMockWorker
from tests.cpe_match.worker.mock_producer import (
    generate_cpe_name,
    uuid_replace,
    uuid_replace_str,
)


def parse_worker_args(raw_args):
    parser = argparse.ArgumentParser()
    CpeMatchProcessor.add_args_to_parser(
        parser
    )  # for common args like "verbose"

    CpeMatchDatabaseProducer.add_args_to_parser(parser)
    return parser.parse_args(raw_args)


class ParseArgsTestCase(unittest.TestCase):
    @patch(
        "greenbone.scap.cpe_match.producer.db.CpeMatchDatabaseProducer",
        autospec=True,
    )
    def test_defaults(self, mock_worker_init: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        args = parse_worker_args([])
        CpeMatchDatabaseProducer.from_args(
            args, console, error_console, progress
        )

        mock_worker_init.assert_called_once_with(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name=None,
            database_schema=None,
            database_host=None,
            database_port=None,
            database_user=None,
            database_password=None,
            echo_sql=False,
            verbose=DEFAULT_VERBOSITY,
        )

    @patch(
        "greenbone.scap.cpe_match.producer.db.CpeMatchDatabaseProducer",
        autospec=True,
    )
    def test_database(self, mock_worker_init: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        args = parse_worker_args(
            [
                "--database-name",
                "test-db-name",
                "--database-schema",
                "test-db-schema",
                "--database-host",
                "test-db-host",
                "--database-port",
                "12345",
                "--database-user",
                "test-db-user",
                "--database-password",
                "test-db-password",
            ]
        )
        CpeMatchDatabaseProducer.from_args(
            args, console, error_console, progress
        )

        mock_worker_init.assert_called_once_with(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name="test-db-name",
            database_schema="test-db-schema",
            database_host="test-db-host",
            database_port=12345,
            database_user="test-db-user",
            database_password="test-db-password",
            echo_sql=False,
            verbose=DEFAULT_VERBOSITY,
        )

    @patch(
        "greenbone.scap.cpe_match.producer.db.CpeMatchDatabaseProducer",
        autospec=True,
    )
    def test_echo_sql(self, mock_worker_init: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        args = parse_worker_args(["--echo-sql"])
        CpeMatchDatabaseProducer.from_args(
            args, console, error_console, progress
        )

        mock_worker_init.assert_called_once_with(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name=None,
            database_schema=None,
            database_host=None,
            database_port=None,
            database_user=None,
            database_password=None,
            echo_sql=True,
            verbose=DEFAULT_VERBOSITY,
        )


class InitTestCase(unittest.TestCase):
    def test_missing_user(self):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        with self.assertRaises(CLIError):
            CpeMatchDatabaseProducer(
                console=console,
                error_console=error_console,
                progress=progress,
                database_name=None,
                database_schema=None,
                database_host=None,
                database_port=None,
                database_user=None,
                database_password=None,
            )

    def test_missing_password(self):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        with self.assertRaises(CLIError):
            CpeMatchDatabaseProducer(
                console=console,
                error_console=error_console,
                progress=progress,
                database_name=None,
                database_schema=None,
                database_host=None,
                database_port=None,
                database_user="db-test-user",
                database_password=None,
            )

    @patch(
        "greenbone.scap.generic_cli.producer.db.PostgresDatabase", autospec=True
    )
    def test_database_defaults(self, db_mock: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        CpeMatchDatabaseProducer(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name=None,
            database_schema=None,
            database_host=None,
            database_port=None,
            database_user="db-test-user",
            database_password="db-test-password",
        )

        db_mock.assert_called_with(
            user="db-test-user",
            password="db-test-password",
            host="localhost",
            port=5432,
            dbname="scap",
            schema=None,
            echo=False,
        )

    @patch(
        "greenbone.scap.generic_cli.producer.db.PostgresDatabase", autospec=True
    )
    def test_database_custom(self, db_mock: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        CpeMatchDatabaseProducer(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name="db-test-name",
            database_schema="db-test-schema",
            database_host="db-test-host",
            database_port=12345,
            database_user="db-test-user",
            database_password="db-test-password",
        )

        db_mock.assert_called_with(
            user="db-test-user",
            password="db-test-password",
            host="db-test-host",
            port=12345,
            dbname="db-test-name",
            schema="db-test-schema",
            echo=False,
        )

    @patch(
        "greenbone.scap.generic_cli.producer.db.PostgresDatabase", autospec=True
    )
    def test_echo_sql(self, db_mock: MagicMock):
        console = Console()
        error_console = Console()
        progress = Progress()

        CpeMatchDatabaseProducer(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name=None,
            database_schema=None,
            database_host=None,
            database_port=None,
            database_user="db-test-user",
            database_password="db-test-password",
            echo_sql=True,
        )

        db_mock.assert_called_with(
            user="db-test-user",
            password="db-test-password",
            host="localhost",
            port=5432,
            dbname="scap",
            schema=None,
            echo=True,
        )


async def async_generate_db_cpe_match_strings(
    num_match_strings: int,
    base_match_criteria_id: UUID,
    base_cpe_name_id: UUID,
):
    for i in range(num_match_strings):
        match_criteria_id = uuid_replace_str(base_match_criteria_id, 1, i)
        cpe_name = generate_cpe_name(1, i)
        cpe_name_id = uuid_replace_str(base_cpe_name_id, 1, i)

        now = datetime.now()
        new_match_model = CPEMatchDatabaseModel()
        new_match_model.cpe_name = cpe_name
        new_match_model.cpe_name_id = cpe_name_id

        new_model = CPEMatchStringDatabaseModel()
        new_model.match_criteria_id = match_criteria_id
        new_model.criteria = cpe_name
        new_model.last_modified = now
        new_model.cpe_last_modified = now - timedelta(days=-1)
        new_model.created = now - timedelta(days=-2)
        new_model.status = "Active"
        new_model.matches = [new_match_model]

        yield new_model


class ProduceTestCase(unittest.IsolatedAsyncioTestCase):
    NUM_CHUNKS = 5
    CHUNK_SIZE = 3
    QUEUE_SIZE = 2

    @patch(
        "greenbone.scap.generic_cli.producer.db.PostgresDatabase", autospec=True
    )
    @patch.object(CPEMatchStringDatabaseManager, "count")
    @patch.object(CPEMatchStringDatabaseManager, "all")
    async def test_read_from_db(
        self, all_mock: AsyncMock, count_mock: AsyncMock, db_mock: AsyncMock
    ):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        producer = CpeMatchDatabaseProducer(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name="test-db-name",
            database_schema="test-db-name",
            database_host="test-db-host",
            database_port=12345,
            database_user="test-db-user",
            database_password="test-db-password",
        )
        worker = CpeMatchMockWorker(
            console=console,
            error_console=error_console,
            progress=progress,
        )
        processor = CpeMatchProcessor(
            console,
            error_console,
            producer,
            worker,
            queue_size=self.QUEUE_SIZE,
            chunk_size=self.CHUNK_SIZE,
        )
        EXPECTED_COUNT = self.NUM_CHUNKS * self.CHUNK_SIZE
        count_mock.return_value = EXPECTED_COUNT

        base_match_criteria_id = uuid4()
        base_cpe_name_id = uuid4()
        all_mock.return_value = async_generate_db_cpe_match_strings(
            EXPECTED_COUNT,
            base_match_criteria_id,
            base_cpe_name_id,
        )

        async with asyncio.timeout(10):
            await processor.run()

            db_mock.assert_called_once_with(
                dbname="test-db-name",
                schema="test-db-name",
                host="test-db-host",
                port=12345,
                user="test-db-user",
                password="test-db-password",
                echo=False,
            )

        self.assertEqual(EXPECTED_COUNT, worker.item_count)

        for i in range(EXPECTED_COUNT):
            expected_match_criteria_id = uuid_replace(
                base_match_criteria_id, 1, i
            )
            expected_cpe_name_id = uuid_replace(base_cpe_name_id, 1, i)
            expected_cpe_name = generate_cpe_name(1, i)

            match_string = worker.items_received[i]
            self.assertEqual(
                expected_match_criteria_id, match_string.match_criteria_id
            )
            self.assertIsInstance(expected_cpe_name_id, UUID)
            self.assertEqual(expected_cpe_name, match_string.criteria)
            self.assertIsInstance(match_string.created, datetime)
            self.assertIsInstance(match_string.last_modified, datetime)
            self.assertIsInstance(match_string.cpe_last_modified, datetime)
            self.assertEqual(
                expected_cpe_name_id, match_string.matches[0].cpe_name_id
            )
            self.assertIsInstance(match_string.matches[0].cpe_name_id, UUID)
            self.assertEqual(
                expected_cpe_name, match_string.matches[0].cpe_name
            )



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

from argparse import Namespace
from typing import AsyncIterator
from uuid import UUID

from pontos.nvd.models.cpe_match_string import CPEMatch, CPEMatchString
from rich.console import Console
from rich.progress import Progress

from greenbone.scap.cli import DEFAULT_VERBOSITY
from greenbone.scap.cpe_match.cli.processor import CPE_MATCH_TYPE_PLURAL
from greenbone.scap.cpe_match.db.manager import CPEMatchStringDatabaseManager
from greenbone.scap.cpe_match.db.models import (
    BaseDatabaseModel,
    CPEMatchStringDatabaseModel,
)
from greenbone.scap.errors import ScapError
from greenbone.scap.generic_cli.producer.db import DatabaseProducer


class CpeMatchDatabaseProducer(DatabaseProducer[CPEMatchString]):
    """
    Async context manager class for a producer querying
    CPE match strings from a database.
    """

    _item_type_plural = CPE_MATCH_TYPE_PLURAL
    "Plural form of the type of items to use in log messages"

    _arg_defaults = DatabaseProducer._arg_defaults
    "Default values for optional arguments."

    @classmethod
    def from_args(
        cls,
        args: Namespace,
        console: Console,
        error_console: Console,
        progress: Progress,
    ) -> "CpeMatchDatabaseProducer":
        """
        Create a new `CPEMatchDatabaseProducer` with parameters from
         the given command line args gathered by an `ArgumentParser`.

        Args:
            args: Command line arguments to use
            console: Console for standard output.
            error_console: Console for error output.
            progress: Progress bar renderer to be updated by the producer.

        Returns:
            The new `CpeMatchDatabaseProducer`.
        """
        return CpeMatchDatabaseProducer(
            console,
            error_console,
            progress,
            database_name=args.database_name,
            database_schema=args.database_schema,
            database_host=args.database_host,
            database_port=args.database_port,
            database_user=args.database_user,
            database_password=args.database_password,
            echo_sql=args.echo_sql,
            verbose=args.verbose or 0,
        )

    def __init__(
        self,
        console: Console,
        error_console: Console,
        progress: Progress,
        *,
        database_name: str,
        database_schema: str,
        database_host: str,
        database_port: int,
        database_user: str,
        database_password: str,
        echo_sql: bool = False,
        verbose: int = DEFAULT_VERBOSITY,
    ):
        """
        Constructor for a CPE match string database producer.

        If the `database_...` arguments are None or not given, corresponding
        environment variables will be tried next before finally using the
        defaults as a fallback.

        Args:
            console: Console for standard output.
            error_console: Console for error output.
            progress: Progress bar renderer to be updated by the producer.
            database_name: Name of the database to use.
            database_schema: Optional database schema to use.
            database_host: IP address or hostname of the database server to use.
            database_port: Port of the database server to use.
            database_user: Name of the database user to use.
            database_password: Password of the database user to use.
            echo_sql: Whether to print SQL statements.
            verbose: Verbosity level of log messages.
        """
        self._manager: CPEMatchStringDatabaseManager

        super().__init__(
            console,
            error_console,
            progress,
            database_name=database_name,
            database_schema=database_schema,
            database_host=database_host,
            database_port=database_port,
            database_user=database_user,
            database_password=database_password,
            echo_sql=echo_sql,
            verbose=verbose,
        )

    def _create_manager(self) -> CPEMatchStringDatabaseManager:
        """
        Callback creating a new database manager for handling CPE match strings.

        Returns: The new database manager.
        """
        return CPEMatchStringDatabaseManager(self._database)

    def _convert_db_model(self, db_model: BaseDatabaseModel) -> CPEMatchString:
        """
        Callback converting a CPE match string database model to a Pontos model.

        Args:
            db_model: The database model convert

        Returns:
            The converted model object.
        """
        if not isinstance(db_model, CPEMatchStringDatabaseModel):
            raise ScapError(
                f"DB model is not a CPEMatchDatabaseModel: {db_model}"
            )
        match_string_db_model: CPEMatchStringDatabaseModel = db_model

        if not match_string_db_model.match_criteria_id:
            raise ScapError(f"Missing match_criteria_id in {db_model}")
        if match_string_db_model.matches is None:
            raise ScapError(f"Missing matches in {db_model}")

        matches = []
        for db_match in match_string_db_model.matches:
            if not db_match.cpe_name_id:
                raise ScapError(f"Missing cpe_name_id in {db_match}")

            matches.append(
                CPEMatch(
                    cpe_name_id=UUID(str(db_match.cpe_name_id)),
                    cpe_name=db_match.cpe_name,
                )
            )

        return CPEMatchString(
            match_criteria_id=UUID(
                str(match_string_db_model.match_criteria_id)
            ),
            criteria=match_string_db_model.criteria,
            status=match_string_db_model.status,
            cpe_last_modified=match_string_db_model.cpe_last_modified,
            created=match_string_db_model.created,
            last_modified=match_string_db_model.last_modified,
            matches=matches,
            version_start_including=match_string_db_model.version_start_including,
            version_start_excluding=match_string_db_model.version_start_excluding,
            version_end_including=match_string_db_model.version_end_including,
            version_end_excluding=match_string_db_model.version_end_excluding,
        )

    async def _db_item_count(self) -> int:
        """
        Callback getting the total number of CPE match strings in the database.

        Returns:
            The total number of CPE match strings
        """
        return await self._manager.count()

    def _db_item_iter(self) -> AsyncIterator[BaseDatabaseModel]:
        """
        Callback getting an async iterator of database items to process.

        Returns:
            An async iterator over database items.
        """
        return self._manager.all()



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gzip
from dataclasses import asdict, dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Optional, Sequence, TextIO

from pontos.nvd.models.cpe_match_string import CPEMatchString
from rich.console import Console

from greenbone.scap.data_utils.json import (
    JsonEncoder,
    JsonManager,
    convert_keys_to_camel,
)


@dataclass
class MatchStringItem:
    """
    Data class to represent a CPE match string item.

    Attributes:
        match_string: The match string instance associated with this item.
    """

    match_string: CPEMatchString


@dataclass
class MatchStringResponse:
    """
    Data class to represent a response containing CPE match string data.

    Attributes:
        results_per_page: Number of results per page.
        start_index: The starting index of the results.
        total_results: Total number of results.
        timestamp: Timestamp of the response creation.
        products: List of products.
        format: Format of the response, default is "NVD_CPE".
        version: Version of the response format, default is "2.0".
    """

    results_per_page: int
    start_index: int
    total_results: int
    timestamp: datetime

    format: str = "NVD_CPEMatchString"
    version: str = "2.0"
    match_strings: list[MatchStringItem] = field(default_factory=list)


class MatchStringJsonManager(JsonManager):
    """
    Manages the storage and organization of CPE match data.
    """

    def __init__(
        self,
        error_console: Console,
        storage_path: Path,
        *,
        compress: bool = False,
        schema_path: Path | None = None,
        raise_error_on_validation=False,
    ):
        super().__init__(
            error_console=error_console,
            schema_path=schema_path,
            raise_error_on_validation=raise_error_on_validation,
        )
        self._match_string_response = MatchStringResponse(
            results_per_page=1,
            start_index=0,
            total_results=1,
            timestamp=datetime.now(),
            match_strings=[],
        )
        self._compress: bool = compress
        self._storage_path: Path = storage_path

    def add_match_string(self, match_string: CPEMatchString) -> None:
        self._match_string_response.match_strings.append(
            MatchStringItem(match_string=match_string)
        )

    def add_match_strings(
        self, match_strings: Sequence[CPEMatchString]
    ) -> None:
        for match_string in match_strings:
            self.add_match_string(match_string)

    def _encode_json(
        self,
        data: dict[str, Any],
        out_file: TextIO,
        validation_buffer: Optional[bytearray] = None,
        *,
        indent: int = 1,
    ):
        encoder = JsonEncoder(indent=indent)

        for chunk in encoder.iterencode(data):
            out_file.write(chunk)
            if validation_buffer is not None:
                validation_buffer.extend(chunk.encode("utf-8"))

    def write(self, file_name: str = "nvd-cpe-matches") -> None:
        """
        Write the CPE data to JSON files with optional compression in the specified folder.
        """

        self._match_string_response.results_per_page = len(
            self._match_string_response.match_strings
        )
        self._match_string_response.total_results = len(
            self._match_string_response.match_strings
        )

        validation_buffer: Optional[bytearray] = None
        if self.validate:
            validation_buffer = bytearray()

        response_dict = asdict(self._match_string_response)
        convert_keys_to_camel(response_dict)

        if self._compress:
            path = self._storage_path / f"{file_name}.json.gz"
            with gzip.open(path, "wt", encoding="utf-8") as out_file:
                self._encode_json(response_dict, out_file, validation_buffer)
        else:
            path = self._storage_path / f"{file_name}.json"
            with open(path, "wt", encoding="utf-8") as out_file:
                self._encode_json(response_dict, out_file, validation_buffer)

        if validation_buffer:
            self._validate_json(file_name, validation_buffer.decode("utf-8"))



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
import os
from argparse import ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path
from typing import Sequence

import shtab
import stamina
from pontos.nvd import now
from pontos.nvd.api import NVDResults
from pontos.nvd.cpe import CPEApi
from pontos.nvd.models.cpe import CPE
from rich.console import Console
from rich.progress import Progress

from greenbone.scap.cli import (
    DEFAULT_POSTGRES_DATABASE_NAME,
    DEFAULT_POSTGRES_HOST,
    DEFAULT_POSTGRES_PORT,
    DEFAULT_POSTGRES_USER,
    DEFAULT_RETRIES,
    DEFAULT_VERBOSITY,
    CLIError,
    CLIRunner,
)
from greenbone.scap.constants import STAMINA_API_RETRY_EXCEPTIONS
from greenbone.scap.cpe.manager import CPEManager
from greenbone.scap.db import PostgresDatabase
from greenbone.scap.timer import Timer

# disable stamina logging
stamina.instrumentation.set_on_retry_hooks([])


DEFAULT_QUEUE_SIZE = 3


def parse_args(args: Sequence[str] | None = None) -> Namespace:
    parser = ArgumentParser(
        description="Create and update a CPE database. "
        "Downloads CPE information from the NIST NVD REST API into the database. "
    )
    shtab.add_argument_to(parser)

    db_group = parser.add_argument_group()
    db_group.add_argument(
        "--database-name",
        help="Name of the CPE database.",
    )
    db_group.add_argument(
        "--database-host",
        help="Name of the CPE database host.",
    )
    db_group.add_argument(
        "--database-port",
        help="Name of the CPE database port.",
        type=int,
    )
    db_group.add_argument(
        "--database-user",
        help="Name of the CPE database user.",
    )
    db_group.add_argument(
        "--database-password",
        help="Name of the CPE database password.",
    )
    db_group.add_argument(
        "--database-schema",
        help="Name of the CPE database schema.",
    )

    since_group = parser.add_mutually_exclusive_group()
    since_group.add_argument(
        "--since",
        metavar="DATE",
        type=datetime.fromisoformat,
        help="Load all changes since a specific date",
    )
    since_group.add_argument(
        "--since-from-file",
        type=Path,
        metavar="FILE",
        help="Load all changes since a specific date. The date is read from "
        "FILE.",
    )

    parser.add_argument(
        "--store-runtime",
        metavar="FILE",
        type=Path,
        help="Store time of this run in FILE",
    )

    parser.add_argument(
        "--echo-sql",
        action="store_true",
        help="Print out all SQL queries.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Enable verbose output.",
    )
    parser.add_argument(
        "--number",
        "-n",
        metavar="N",
        help="Fetch up to N CPEs only",
        type=int,
    )
    parser.add_argument(
        "--retry-attempts",
        type=int,
        metavar="N",
        help="Up to N retries until giving up when HTTP requests are failing. "
        f"Default: {DEFAULT_RETRIES}",
    )
    parser.add_argument(
        "--nvd-api-key",
        metavar="KEY",
        help="Use a NVD API key for downloading the CPEs. Using an API key "
        "allows for downloading with extended rate limits.",
    )
    parser.add_argument(
        "--chunk-size",
        help="Number of CPEs to download and process in one request. A lower "
        "number allows for more frequent updates and feedback.",
        type=int,
        metavar="N",
    )
    parser.add_argument(
        "--queue-size",
        help="Size of the download queue. It sets the maximum number of CPEs "
        "kept in the memory. The maximum number of CPEs is chunk size * queue "
        "size. Default: %(default)s.",
        type=int,
        metavar="N",
        default=DEFAULT_QUEUE_SIZE,
    )

    return parser.parse_args(args)


class CPECli:
    def __init__(
        self,
        console: Console,
        *,
        verbose: int = 0,
        chunk_size: int | None = None,
        queue_size: int = DEFAULT_QUEUE_SIZE,
    ) -> None:
        self.queue: asyncio.Queue[Sequence[CPE]] = asyncio.Queue(queue_size)
        self.chunk_size = chunk_size
        self.console = console
        self.event = asyncio.Event()
        self.verbose = verbose

    async def _worker(
        self,
        progress: Progress,
        manager: CPEManager,
        total_cpes: int,
    ) -> None:
        self.console.log("Start processing CPEs")
        processed = 0

        task = progress.add_task("Processing CPEs", total=total_cpes)

        while not self.event.is_set() or not self.queue.empty():
            try:
                cpes = await self.queue.get()
                processed += len(cpes)

                progress.update(task, completed=processed)

                await manager.add_cpes(cpes)
            except asyncio.CancelledError as e:
                if self.verbose:
                    self.console.log("Worker has been cancelled")
                raise e

            self.queue.task_done()

            if self.verbose:
                self.console.log(f"Processed {processed:,} CPEs")

        self.console.log(f"Processing of {processed:,} CPEs done")

    async def _producer(
        self,
        progress: Progress,
        results: NVDResults[CPE],
        retry_attempts: int,
        total_cpes: int,
    ) -> None:
        task = progress.add_task("Downloading CPEs", total=total_cpes)

        try:
            with Timer() as download_timer:
                count = 0
                async for attempt in stamina.retry_context(
                    on=STAMINA_API_RETRY_EXCEPTIONS,
                    attempts=retry_attempts,
                    timeout=None,
                ):
                    with attempt:
                        attempt_number = attempt.num
                        if attempt_number > 1:
                            self.console.log(
                                "HTTP request failed. Download attempt "
                                f"{attempt_number} of {retry_attempts}"
                            )

                        async for cpes in results.chunks():
                            count += len(cpes)
                            progress.update(task, completed=count)

                            if self.verbose:
                                self.console.log(f"Downloaded {count:,} CPEs")

                            await self.queue.put(cpes)

            self.console.log(
                f"Downloaded {count:,} CPEs in "
                f"{download_timer.elapsed_time:0.4f} seconds."
            )

        finally:
            # signal worker that we are finished with downloading
            self.event.set()

    async def _join(self):
        await self.queue.join()

    async def download(
        self,
        progress: Progress,
        manager: CPEManager,
        api: CPEApi,
        retry_attempts: int,
        request_results: int | None,
        last_modified_start_date: datetime | None,
        last_modified_end_date: datetime | None,
    ) -> None:
        async for attempt in stamina.retry_context(
            on=STAMINA_API_RETRY_EXCEPTIONS,
            attempts=retry_attempts,
            timeout=None,
        ):
            with attempt:
                attempt_number = attempt.num
                additional_retry_attempts = retry_attempts - (
                    attempt_number - 1
                )
                if attempt_number > 1:
                    self.console.log(
                        "HTTP request failed. Download attempt "
                        f"{attempt_number} of {retry_attempts}"
                    )
                else:
                    self.console.log(
                        f"Download attempt {attempt_number} of {retry_attempts}"
                    )

                results = await api.cpes(
                    request_results=request_results,
                    last_modified_start_date=last_modified_start_date,
                    last_modified_end_date=last_modified_end_date,
                    results_per_page=self.chunk_size,
                )

        result_count = len(results)  # type: ignore

        self.console.log(f"{result_count:,} CPEs to download available")

        if request_results == 0 or not result_count:
            # no new CPEs available or no CPEs requested
            return

        total_cpes = min(request_results or result_count, result_count)

        async with asyncio.TaskGroup() as tg:
            tg.create_task(
                self._worker(
                    progress,
                    manager,
                    total_cpes=total_cpes,
                )
            )
            producer = tg.create_task(
                self._producer(
                    progress,
                    results,
                    additional_retry_attempts,
                    total_cpes=total_cpes,
                )
            )
            await producer
            await self._join()


async def download(console: Console, error_console: Console) -> None:
    args = parse_args()

    since: datetime | None = args.since
    retry_attempts: int = args.retry_attempts or int(
        os.environ.get("RETRY_ATTEMPTS", DEFAULT_RETRIES)
    )
    echo_sql: bool = args.echo_sql
    verbose: int = (
        args.verbose
        if args.verbose is not None
        else int(os.environ.get("VERBOSE", DEFAULT_VERBOSITY))
    )
    since_from_file: Path | None = args.since_from_file
    run_time_file: Path | None = args.store_runtime
    until = now() if run_time_file else None
    number: int | None = args.number
    nvd_api_key: str | None = args.nvd_api_key or os.environ.get("NVD_API_KEY")

    chunk_size: int | None = args.chunk_size
    queue_size: int = args.queue_size

    if since_from_file:
        if since_from_file.exists():
            since = datetime.fromisoformat(
                since_from_file.read_text(encoding="utf8").strip()
            )
        else:
            error_console.print(
                f"{since_from_file.absolute()} does not exist. Ignoring "
                "--since-from-file argument."
            )
            since = None

    cpe_database_name: str = (
        args.database_name
        or os.environ.get("CPE_DATABASE_NAME")
        or os.environ.get("DATABASE_NAME")
        or DEFAULT_POSTGRES_DATABASE_NAME
    )
    cpe_database_user: str = (
        args.database_user
        or os.environ.get("CPE_DATABASE_USER")
        or os.environ.get("DATABASE_USER")
        or DEFAULT_POSTGRES_USER
    )
    cpe_database_host: str = (
        args.database_host
        or os.environ.get("CPE_DATABASE_HOST")
        or os.environ.get("DATABASE_HOST")
        or DEFAULT_POSTGRES_HOST
    )
    cpe_database_port: int = int(
        args.database_port
        or os.environ.get("CPE_DATABASE_PORT")
        or os.environ.get("DATABASE_PORT")
        or DEFAULT_POSTGRES_PORT
    )
    cpe_database_schema: str | None = (
        args.database_schema
        or os.environ.get("CPE_DATABASE_SCHEMA")
        or os.environ.get("DATABASE_SCHEMA")
    )
    cpe_database_password: str | None = (
        args.database_password
        or os.environ.get("CPE_DATABASE_PASSWORD")
        or os.environ.get("DATABASE_PASSWORD")
    )
    if not cpe_database_password:
        raise CLIError("Missing password for CPE database")

    cpe_database = PostgresDatabase(
        user=cpe_database_user,
        password=cpe_database_password,
        host=cpe_database_host,
        port=cpe_database_port,
        dbname=cpe_database_name,  # type: ignore
        schema=cpe_database_schema,
        echo=echo_sql,
    )
    if verbose:
        console.log(f"Using PostgreSQL database {cpe_database_name}")

    cli = CPECli(
        console, verbose=verbose, chunk_size=chunk_size, queue_size=queue_size
    )

    with Progress(console=console) as progress:
        async with (
            cpe_database,
            CPEApi(token=nvd_api_key) as api,
            CPEManager(cpe_database) as manager,
        ):
            if verbose:
                console.log("Initialized database.")

                if nvd_api_key:
                    console.log("Using NVD API key to download the CPEs")

            console.log("Start downloading CPEs")

            if since:
                console.log(
                    "Downloading changed or new CPEs since "
                    f"{since.isoformat()}"
                )

            await cli.download(
                progress,
                manager,
                api,
                retry_attempts,
                request_results=number,
                last_modified_start_date=since,
                last_modified_end_date=since and until,
            )

            if run_time_file:
                if until:
                    run_time = until
                else:
                    run_time = datetime.now()
                # ensure directories exist
                run_time_file.parent.mkdir(parents=True, exist_ok=True)
                run_time_file.write_text(
                    f"{run_time.isoformat()}\n",
                    encoding="utf8",  # type: ignore
                )
                console.log(f"Wrote run time to {run_time_file.absolute()}.")


def main() -> None:
    CLIRunner.run(download)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import asyncio
import unittest
from unittest import mock
from unittest.mock import AsyncMock, MagicMock, patch

from rich.console import Console
from rich.progress import Progress

from greenbone.scap.cli import DEFAULT_VERBOSITY, CLIError
from greenbone.scap.cpe_match.cli.processor import CpeMatchProcessor
from greenbone.scap.cpe_match.db.models import CPEMatchStringDatabaseModel
from greenbone.scap.cpe_match.worker.db import CpeMatchDatabaseWriteWorker
from tests.cpe_match.worker.mock_producer import CpeMatchMockProducer


def parse_worker_args(raw_args):
    parser = argparse.ArgumentParser()
    CpeMatchProcessor.add_args_to_parser(
        parser
    )  # for common args like "verbose"

    CpeMatchDatabaseWriteWorker.add_args_to_parser(parser)
    return parser.parse_args(raw_args)


class ParseArgsTestCase(unittest.TestCase):
    @patch(
        "greenbone.scap.cpe_match.worker.db.CpeMatchDatabaseWriteWorker",
        autospec=True,
    )
    def test_defaults(self, mock_worker_init: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        args = parse_worker_args([])
        CpeMatchDatabaseWriteWorker.from_args(
            args, console, error_console, progress
        )

        mock_worker_init.assert_called_once_with(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name=None,
            database_schema=None,
            database_host=None,
            database_port=None,
            database_user=None,
            database_password=None,
            echo_sql=False,
            verbose=DEFAULT_VERBOSITY,
        )

    @patch(
        "greenbone.scap.cpe_match.worker.db.CpeMatchDatabaseWriteWorker",
        autospec=True,
    )
    #    @patch("greenbone.scap.db.PostgresDatabase", autospec=True)
    def test_database(self, mock_worker_init: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        args = parse_worker_args(
            [
                "--database-name",
                "test-db-name",
                "--database-schema",
                "test-db-schema",
                "--database-host",
                "test-db-host",
                "--database-port",
                "12345",
                "--database-user",
                "test-db-user",
                "--database-password",
                "test-db-password",
            ]
        )
        CpeMatchDatabaseWriteWorker.from_args(
            args, console, error_console, progress
        )

        mock_worker_init.assert_called_once_with(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name="test-db-name",
            database_schema="test-db-schema",
            database_host="test-db-host",
            database_port=12345,
            database_user="test-db-user",
            database_password="test-db-password",
            echo_sql=False,
            verbose=DEFAULT_VERBOSITY,
        )

    @patch(
        "greenbone.scap.cpe_match.worker.db.CpeMatchDatabaseWriteWorker",
        autospec=True,
    )
    def test_echo_sql(self, mock_worker_init: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        args = parse_worker_args(["--echo-sql"])
        CpeMatchDatabaseWriteWorker.from_args(
            args, console, error_console, progress
        )

        mock_worker_init.assert_called_once_with(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name=None,
            database_schema=None,
            database_host=None,
            database_port=None,
            database_user=None,
            database_password=None,
            echo_sql=True,
            verbose=DEFAULT_VERBOSITY,
        )


class InitTestCase(unittest.TestCase):
    def test_missing_user(self):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        with self.assertRaises(CLIError):
            CpeMatchDatabaseWriteWorker(
                console=console,
                error_console=error_console,
                progress=progress,
                database_name=None,
                database_schema=None,
                database_host=None,
                database_port=None,
                database_user=None,
                database_password=None,
            )

    def test_missing_password(self):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        with self.assertRaises(CLIError):
            CpeMatchDatabaseWriteWorker(
                console=console,
                error_console=error_console,
                progress=progress,
                database_name=None,
                database_schema=None,
                database_host=None,
                database_port=None,
                database_user="db-test-user",
                database_password=None,
            )

    @patch(
        "greenbone.scap.generic_cli.worker.db.PostgresDatabase", autospec=True
    )
    def test_database_defaults(self, db_mock: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        CpeMatchDatabaseWriteWorker(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name=None,
            database_schema=None,
            database_host=None,
            database_port=None,
            database_user="db-test-user",
            database_password="db-test-password",
        )

        db_mock.assert_called_with(
            user="db-test-user",
            password="db-test-password",
            host="localhost",
            port=5432,
            dbname="scap",
            schema=None,
            echo=False,
        )

    @patch(
        "greenbone.scap.generic_cli.worker.db.PostgresDatabase", autospec=True
    )
    def test_database_custom(self, db_mock: MagicMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        CpeMatchDatabaseWriteWorker(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name="db-test-name",
            database_schema="db-test-schema",
            database_host="db-test-host",
            database_port=12345,
            database_user="db-test-user",
            database_password="db-test-password",
        )

        db_mock.assert_called_with(
            user="db-test-user",
            password="db-test-password",
            host="db-test-host",
            port=12345,
            dbname="db-test-name",
            schema="db-test-schema",
            echo=False,
        )

    @patch(
        "greenbone.scap.generic_cli.worker.db.PostgresDatabase", autospec=True
    )
    def test_echo_sql(self, db_mock: MagicMock):
        console = Console()
        error_console = Console()
        progress = Progress()

        CpeMatchDatabaseWriteWorker(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name=None,
            database_schema=None,
            database_host=None,
            database_port=None,
            database_user="db-test-user",
            database_password="db-test-password",
            echo_sql=True,
        )

        db_mock.assert_called_with(
            user="db-test-user",
            password="db-test-password",
            host="localhost",
            port=5432,
            dbname="scap",
            schema=None,
            echo=True,
        )


class WriteTestCase(unittest.IsolatedAsyncioTestCase):
    NUM_CHUNKS = 5
    CHUNK_SIZE = 3
    QUEUE_SIZE = 2

    @patch(
        "greenbone.scap.generic_cli.worker.db.PostgresDatabase", autospec=True
    )
    async def test_write_json(self, db_mock: AsyncMock):
        console = Console(quiet=True)
        error_console = Console(quiet=True)
        progress = Progress(disable=True)

        producer = CpeMatchMockProducer(
            console=console,
            error_console=error_console,
            progress=progress,
            num_chunks=self.NUM_CHUNKS,
            chunk_size=self.CHUNK_SIZE,
        )
        worker = CpeMatchDatabaseWriteWorker(
            console=console,
            error_console=error_console,
            progress=progress,
            database_name="test-db-name",
            database_schema="test-db-name",
            database_host="test-db-host",
            database_port=12345,
            database_user="test-db-user",
            database_password="test-db-password",
        )
        processor = CpeMatchProcessor(
            console,
            error_console,
            producer,
            worker,
            queue_size=self.QUEUE_SIZE,
            chunk_size=self.CHUNK_SIZE,
        )
        db_mock.insert.side_effect = lambda x: print("x=", x)

        async with asyncio.timeout(10):
            await processor.run()

            db_mock.assert_called_once_with(
                dbname="test-db-name",
                schema="test-db-name",
                host="test-db-host",
                port=12345,
                user="test-db-user",
                password="test-db-password",
                echo=False,
            )
            # check if there is at least one insert call for each chunk
            insert_call = mock.call().insert(CPEMatchStringDatabaseModel)
            self.assertGreaterEqual(
                self.NUM_CHUNKS, db_mock.mock_calls.count(insert_call)
            )



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC
from argparse import ArgumentParser
from pathlib import Path
from typing import Type, TypeVar

from rich.console import Console
from rich.progress import Progress

from ...cli import DEFAULT_VERBOSITY
from .base import BaseScapWorker

T = TypeVar("T")
"Generic type variable for the type of SCAP items handled"


class ScapJsonWriteWorker(BaseScapWorker[T], ABC):
    """
    Abstract async context manager base class for a worker writing
    SCAP items to a single JSON file.

    The type of the SCAP items is to be specified by the generic type,
    e.g. `ScapJsonWriteWorker[CPE]` will be a producer handling CPE objects.
    """

    _item_type_plural = BaseScapWorker._item_type_plural
    "Default values for optional arguments."

    _arg_defaults = {
        "storage_path": ".",
        "schema_path": None,
        "verbose": DEFAULT_VERBOSITY,
    }
    "Default values for optional arguments."

    @classmethod
    def add_args_to_parser(
        cls: Type["ScapJsonWriteWorker"],
        parser: ArgumentParser,
    ):
        """
        Class method for adding JSON writer arguments to an
         `ArgumentParser`.

        Args:
            parser: The parser to add the arguments to.
        """
        parser.add_argument(
            "--storage-path",
            type=Path,
            help="Directory to write the JSON to. Default: %(default)s",
            default=cls._arg_defaults["storage_path"],
        )
        parser.add_argument(
            "--compress",
            action="store_true",
            default=False,
            help="Gzip compress the resulting JSON file.",
        )
        parser.add_argument(
            "--schema-path",
            default=cls._arg_defaults["schema_path"],
            type=Path,
            help="Json schema file path. Default: %(default)s.",
        )

    def __init__(
        self,
        console: Console,
        error_console: Console,
        progress: Progress,
        *,
        storage_path: Path,
        schema_path: Path | None = None,
        compress: bool = False,
        verbose: int | None = None,
    ):
        """
        Constructor for a `ScapJsonWriteWorker`.

        Args:
            console: Console for standard output.
            error_console: Console for error output.
            progress: Progress bar renderer to be updated by the producer.
            storage_path: Path to the directory to write the JSON file into.
            schema_path: Optional path to the schema file for JSON validation.
            compress: Whether to gzip compress the JSON file.
            verbose: Verbosity level of log messages.
        """
        super().__init__(console, error_console, progress, verbose=verbose)

        self._storage_path = storage_path
        "Path to the directory to write the JSON file into."

        self._schema_path = schema_path
        "Optional path to the schema file for JSON validation."

        self._compress = compress
        "Whether to gzip compress the JSON file."

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
import os
import sys
from argparse import ArgumentParser, BooleanOptionalAction, Namespace
from typing import Sequence

import shtab
from pontos.cpe import CPE
from rich.console import Console

from greenbone.scap.cli import (
    DEFAULT_POSTGRES_DATABASE_NAME,
    DEFAULT_POSTGRES_HOST,
    DEFAULT_POSTGRES_PORT,
    DEFAULT_POSTGRES_USER,
)
from greenbone.scap.cpe.manager import CPEManager, VersionRange
from greenbone.scap.db import PostgresDatabase
from greenbone.scap.errors import ScapError


def parse_args(args: Sequence[str] | None = None) -> Namespace:
    parser = ArgumentParser(description="Search for CPEs in the database")
    shtab.add_argument_to(parser)

    parser.add_argument("cpe", help="CPE match to search for")
    parser.add_argument(
        "--echo-sql",
        action="store_true",
        help="Print out all SQL queries.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Enable verbose output.",
    )
    parser.add_argument(
        "--exact",
        action="store_true",
        help="Find only CPE information that match exactly the provided CPE "
        "name",
    )
    db_group = parser.add_argument_group()
    db_group.add_argument(
        "--database-name",
        help="Name of the CPE database.",
    )
    db_group.add_argument(
        "--database-host",
        help="Name of the CPE database host.",
    )
    db_group.add_argument(
        "--database-port",
        help="Name of the CPE database port.",
        type=int,
    )
    db_group.add_argument(
        "--database-user",
        help="Name of the CPE database user.",
    )
    db_group.add_argument(
        "--database-password",
        help="Name of the CPE database password.",
    )
    db_group.add_argument(
        "--database-schema",
        help="Name of the CPE database schema.",
    )

    filter_group = parser.add_argument_group("Filter")
    filter_group.add_argument(
        "--version-start-including",
        metavar="VERSION",
        help="Find CPEs with version greater or equal then VERSION",
    )
    filter_group.add_argument(
        "--version-start-excluding",
        metavar="VERSION",
        help="Find CPEs with version greater then VERSION",
    )
    filter_group.add_argument(
        "--version-end-including",
        metavar="VERSION",
        help="Find CPEs with version less or equal then VERSION",
    )
    filter_group.add_argument(
        "--version-end-excluding",
        metavar="VERSION",
        help="Find CPEs with version less then VERSION",
    )
    filter_group.add_argument(
        "--limit",
        type=int,
        metavar="N",
        help="Limit to N number of CPEs",
    )
    filter_group.add_argument(
        "--include-deprecated",
        action=BooleanOptionalAction,
        default=False,
        help="Consider deprecated CPEs",
    )
    return parser.parse_args(args)


async def find_cpes(console: Console) -> None:
    args = parse_args()

    echo_sql: bool = args.echo_sql
    verbose: int = args.verbose
    version_start_including: str | None = args.version_start_including
    version_start_excluding: str | None = args.version_start_excluding
    version_end_including: str | None = args.version_end_including
    version_end_excluding: str | None = args.version_end_excluding
    limit: int | None = args.limit
    include_deprecated: bool = args.include_deprecated

    cpe_database_name: str = (
        args.database_name
        or os.environ.get("CPE_DATABASE_NAME")
        or os.environ.get("DATABASE_NAME")
        or DEFAULT_POSTGRES_DATABASE_NAME
    )
    cpe_database_user: str = (
        args.database_user
        or os.environ.get("CPE_DATABASE_USER")
        or os.environ.get("DATABASE_USER")
        or DEFAULT_POSTGRES_USER
    )
    cpe_database_host: str = (
        args.database_host
        or os.environ.get("CPE_DATABASE_HOST")
        or os.environ.get("DATABASE_HOST")
        or DEFAULT_POSTGRES_HOST
    )
    cpe_database_port: int = int(
        args.database_port
        or os.environ.get("CPE_DATABASE_PORT")
        or os.environ.get("DATABASE_PORT")
        or DEFAULT_POSTGRES_PORT
    )
    cpe_database_schema: str | None = (
        args.database_schema
        or os.environ.get("CPE_DATABASE_SCHEMA")
        or os.environ.get("DATABASE_SCHEMA")
    )
    cpe_database_password: str | None = (
        args.database_password
        or os.environ.get("CPE_DATABASE_PASSWORD")
        or os.environ.get("DATABASE_PASSWORD")
    )
    if not cpe_database_password:
        raise ScapError("Missing password for CPE database")

    cpe_database = PostgresDatabase(
        user=cpe_database_user,
        password=cpe_database_password,
        host=cpe_database_host,
        port=cpe_database_port,
        dbname=cpe_database_name,  # type: ignore
        schema=cpe_database_schema,
        echo=echo_sql,
    )
    if verbose:
        console.log(f"Using PostgreSQL database {cpe_database_name}")

    async with cpe_database as db, CPEManager(db) as manager:
        count = 0
        console.log(f"Searching for CPEs matching '{args.cpe}'")

        async for cpe in manager.find(
            exact=args.exact,
            version_ranges=[
                VersionRange(
                    cpe=CPE.from_string(args.cpe),
                    version_start_excluding=version_start_excluding,
                    version_start_including=version_start_including,
                    version_end_excluding=version_end_excluding,
                    version_end_including=version_end_including,
                )
            ],
            limit=limit,
            deprecated=include_deprecated,
        ):
            count += 1
            console.print(cpe, end="\n\n")

        console.log(f"Found {count} matching CPEs")


def main() -> None:
    console = Console(log_path=False)
    error_console = Console(file=sys.stderr, log_path=False)
    try:
        asyncio.run(find_cpes(console))
    except KeyboardInterrupt:
        pass
    except ScapError as e:
        error_console.print(f"Error: {e}")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
import os
from argparse import ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path
from typing import Sequence

import shtab
import stamina
from pontos.nvd import now
from pontos.nvd.api import NVDResults
from pontos.nvd.cve import CVEApi
from pontos.nvd.models.cve import CVE
from rich.console import Console
from rich.progress import Progress

from greenbone.scap.cli import (
    DEFAULT_POSTGRES_DATABASE_NAME,
    DEFAULT_POSTGRES_HOST,
    DEFAULT_POSTGRES_PORT,
    DEFAULT_POSTGRES_USER,
    DEFAULT_RETRIES,
    DEFAULT_VERBOSITY,
    CLIError,
    CLIRunner,
)
from greenbone.scap.constants import STAMINA_API_RETRY_EXCEPTIONS
from greenbone.scap.cve.manager import CVEManager
from greenbone.scap.db import PostgresDatabase
from greenbone.scap.timer import Timer

# disable stamina logging
stamina.instrumentation.set_on_retry_hooks([])

DEFAULT_QUEUE_SIZE = 10


def parse_args(args: Sequence[str] | None = None) -> Namespace:
    parser = ArgumentParser(
        description="Create and update a CVE database. "
        "Downloads CVE information from the NIST NVD REST API into the database."
    )
    shtab.add_argument_to(parser)

    parser.add_argument(
        "--number",
        "-n",
        metavar="N",
        help="Download up to N CVEs only",
        type=int,
    )

    db_group = parser.add_argument_group(
        title="Database", description="Database related settings"
    )
    db_group.add_argument(
        "--database-name",
        help="Name of the CVE database.",
    )
    db_group.add_argument(
        "--database-host",
        help="Name of the CVE database host.",
    )
    db_group.add_argument(
        "--database-port",
        help="Name of the CVE database port.",
        type=int,
    )
    db_group.add_argument(
        "--database-user",
        help="Name of the CVE database user.",
    )
    db_group.add_argument(
        "--database-password",
        help="Name of the CVE database password.",
    )
    db_group.add_argument(
        "--database-schema",
        help="Name of the CVE database schema.",
    )

    since_group = parser.add_mutually_exclusive_group()
    since_group.add_argument(
        "--since",
        metavar="DATE",
        type=datetime.fromisoformat,
        help="Load all CVE changes since a specific date.",
    )
    since_group.add_argument(
        "--since-from-file",
        type=Path,
        metavar="FILE",
        help="Load all CVE changes since a specific date. The date is read "
        "from FILE.",
    )

    parser.add_argument(
        "--store-runtime",
        metavar="FILE",
        type=Path,
        help="Store time of this run in FILE. It can be used to load only the "
        "modified CVEs next time using the --since-from-time argument.",
    )
    parser.add_argument(
        "--store-updated-cves",
        metavar="FILE",
        type=Path,
        help="Store the list of updated CVEs in FILE. It can be used as "
        "input for the CVE convert tool to only convert updated CVEs.",
    )

    parser.add_argument(
        "--echo-sql",
        action="store_true",
        help="Print out all SQL queries.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Enable verbose output.",
    )
    parser.add_argument(
        "--retry-attempts",
        type=int,
        metavar="N",
        help="Up to N retries until giving up when HTTP requests are failing. "
        f"Default: {DEFAULT_RETRIES}",
    )
    parser.add_argument(
        "--nvd-api-key",
        metavar="KEY",
        help="Use a NVD API key for downloading the CVEs. Using an API key "
        "allows for downloading with extended rate limits.",
    )
    parser.add_argument(
        "--chunk-size",
        help="Number of CVEs to download and process in one request. A lower "
        "number allows for more frequent updates and feedback.",
        type=int,
        metavar="N",
    )
    parser.add_argument(
        "--queue-size",
        help="Size of the download queue. It sets the maximum number of CVEs "
        "kept in the memory. The maximum number of CVEs is chunk size * queue "
        "size. Default: %(default)s.",
        type=int,
        metavar="N",
        default=DEFAULT_QUEUE_SIZE,
    )
    return parser.parse_args(args)


class CVECli:
    def __init__(
        self,
        console: Console,
        *,
        verbose: int = 0,
        chunk_size: int | None = None,
        queue_size: int = DEFAULT_QUEUE_SIZE,
    ) -> None:
        self.queue: asyncio.Queue[Sequence[CVE]] = asyncio.Queue(queue_size)
        self.chunk_size = chunk_size
        self.console = console
        self.event = asyncio.Event()
        self.verbose = verbose
        self.cves_to_update: set[str] = set()

    async def _worker(
        self, progress: Progress, manager: CVEManager, total_cves: int
    ):
        self.console.log("Start processing CVEs")
        processed = 0

        task = progress.add_task("Processing CVEs", total=total_cves)
        while not self.event.is_set() or not self.queue.empty():
            try:
                cves = await self.queue.get()

                await manager.add_cves(cves)

                self.cves_to_update.update((cve.id for cve in cves))

                self.queue.task_done()

                processed += len(cves)
                progress.update(task, completed=processed)

                self.console.log(f"Processed {processed:,} CVEs")
            except asyncio.CancelledError as e:
                self.console.log("Worker has been cancelled")
                raise e

        self.console.log(f"Processing of {processed:,} CVEs done")

    async def _producer(
        self,
        progress: Progress,
        results: NVDResults[CVE],
        retry_attempts: int,
        total_cves: int,
    ) -> None:
        task = progress.add_task("Downloading CVEs", total=total_cves)

        try:
            with Timer() as download_timer:
                count = 0
                async for attempt in stamina.retry_context(
                    on=STAMINA_API_RETRY_EXCEPTIONS,
                    attempts=retry_attempts,
                    timeout=None,
                ):
                    with attempt:
                        attempt_number = attempt.num
                        if attempt_number > 1:
                            self.console.log(
                                "HTTP request failed. Download attempt "
                                f"{attempt_number} of {retry_attempts}"
                            )

                        async for cves in results.chunks():
                            count += len(cves)
                            progress.update(task, completed=count)

                            if self.verbose:
                                self.console.log(f"Downloaded {count:,} CVEs")

                            await self.queue.put(cves)
        finally:
            # signal worker that we are finished with downloading
            self.event.set()

        self.console.log(
            f"Downloaded {count:,} CVEs in "
            f"{download_timer.elapsed_time:0.4f} seconds."
        )

    async def _join(self):
        await self.queue.join()

    async def download(
        self,
        progress: Progress,
        manager: CVEManager,
        api: CVEApi,
        retry_attempts: int,
        request_results: int | None,
        last_modified_start_date: datetime | None,
        last_modified_end_date: datetime | None,
    ) -> None:
        async for attempt in stamina.retry_context(
            on=STAMINA_API_RETRY_EXCEPTIONS,
            attempts=retry_attempts,
            timeout=None,
        ):
            with attempt:
                attempt_number = attempt.num
                additional_retry_attempts = retry_attempts - (
                    attempt_number - 1
                )
                if attempt_number > 1:
                    self.console.log(
                        "HTTP request failed. Download attempt "
                        f"{attempt_number} of {retry_attempts}"
                    )
                else:
                    self.console.log(
                        f"Download attempt {attempt_number} of {retry_attempts}"
                    )

                results = await api.cves(
                    request_results=request_results,
                    last_modified_start_date=last_modified_start_date,
                    last_modified_end_date=last_modified_end_date,
                    results_per_page=self.chunk_size,
                )

        result_count = len(results)  # type: ignore

        self.console.log(f"{result_count:,} CVEs to download available")

        if request_results == 0 or not result_count:
            # no new CVEs available or no CVEs requested
            return

        total_cves = min(request_results or result_count, result_count)

        async with asyncio.TaskGroup() as tg:
            tg.create_task(
                self._worker(
                    progress,
                    manager,
                    total_cves=total_cves,
                )
            )
            producer = tg.create_task(
                self._producer(
                    progress,
                    results,
                    additional_retry_attempts,
                    total_cves=total_cves,
                )
            )
            await producer
            await self._join()


async def download(console: Console, error_console: Console):
    args = parse_args()

    since: datetime | None = args.since
    verbose: int = (
        args.verbose
        if args.verbose is not None
        else int(os.environ.get("VERBOSE", DEFAULT_VERBOSITY))
    )
    run_time_file: Path | None = args.store_runtime
    since_from_file: Path | None = args.since_from_file
    echo_sql: bool = args.echo_sql
    number: int | None = args.number
    retry_attempts: int = args.retry_attempts or int(
        os.environ.get("RETRY_ATTEMPTS", DEFAULT_RETRIES)
    )
    nvd_api_key: str | None = args.nvd_api_key or os.environ.get("NVD_API_KEY")
    updated_cves_file: Path | None = args.store_updated_cves

    chunk_size: int | None = args.chunk_size
    queue_size: int = args.queue_size

    until = now() if run_time_file else None

    if since_from_file:
        if since_from_file.exists():
            since = datetime.fromisoformat(
                since_from_file.read_text(encoding="utf8").strip()
            )
        else:
            error_console.print(
                f"{since_from_file.absolute()} does not exist. Ignoring "
                "--since-from-file argument."
            )
            since = None

    cve_database_name = (
        args.database_name
        or os.environ.get("CVE_DATABASE_NAME")
        or os.environ.get("DATABASE_NAME")
        or DEFAULT_POSTGRES_DATABASE_NAME
    )
    cve_database_user: str = (
        args.database_user
        or os.environ.get("CVE_DATABASE_USER")
        or os.environ.get("DATABASE_USER")
        or DEFAULT_POSTGRES_USER
    )
    cve_database_host: str = (
        args.database_host
        or os.environ.get("CVE_DATABASE_HOST")
        or os.environ.get("DATABASE_HOST")
        or DEFAULT_POSTGRES_HOST
    )
    cve_database_port: int = int(
        args.database_port
        or os.environ.get("CVE_DATABASE_PORT")
        or os.environ.get("DATABASE_PORT")
        or DEFAULT_POSTGRES_PORT
    )
    cve_database_schema: str | None = (
        args.database_schema
        or os.environ.get("CVE_DATABASE_SCHEMA")
        or os.environ.get("DATABASE_SCHEMA")
    )
    cve_database_password: str | None = (
        args.database_password
        or os.environ.get("CVE_DATABASE_PASSWORD")
        or os.environ.get("DATABASE_PASSWORD")
    )
    if not cve_database_password:
        raise CLIError("Missing password for CVE database")

    cve_database = PostgresDatabase(
        user=cve_database_user,
        password=cve_database_password,
        host=cve_database_host,
        port=cve_database_port,
        dbname=cve_database_name,
        schema=cve_database_schema,
        echo=echo_sql,
    )
    if verbose:
        console.log(f"Using PostgreSQL database {cve_database_name} for CVEs")

    cli = CVECli(
        console, verbose=verbose, chunk_size=chunk_size, queue_size=queue_size
    )

    run_time = datetime.now()

    with Progress(console=console) as progress:
        async with (
            cve_database,
            CVEApi(token=nvd_api_key) as api,
            CVEManager(cve_database) as cve_manager,
        ):

            if verbose:
                console.log("Initialized databases.")

                if nvd_api_key:
                    console.log("Using NVD API key to download the CVEs")

            if number != 0:
                console.log("Start downloading CVEs")

                if since:
                    console.log(
                        "Downloading changed or new CVEs since "
                        f"{since.isoformat()}"
                    )

                await cli.download(
                    progress,
                    cve_manager,
                    api,
                    retry_attempts,
                    request_results=number,
                    last_modified_start_date=since,
                    last_modified_end_date=since and until,
                )

        if run_time_file:
            if until:
                run_time = until
            # ensure directories exist
            run_time_file.parent.mkdir(parents=True, exist_ok=True)
            run_time_file.write_text(
                f"{run_time.isoformat()}\n",
                encoding="utf8",  # type: ignore
            )
            console.log(f"Wrote run time to {run_time_file.absolute()}.")

        if updated_cves_file:
            # ensure directories exist
            updated_cves_file.parent.mkdir(parents=True, exist_ok=True)
            updated_cves_file.write_text(
                "\n".join(cli.cves_to_update), encoding="utf8"
            )
            console.log(
                f"Wrote {len(cli.cves_to_update):,} updated CVEs to "
                f"{updated_cves_file.absolute()}."
            )


def main() -> None:
    CLIRunner.run(download)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
from argparse import Namespace

from pontos.nvd import NVDResults
from pontos.nvd.cpe_match import CPEMatchApi
from pontos.nvd.models.cpe_match_string import CPEMatchString
from rich.console import Console
from rich.progress import Progress

from greenbone.scap.cli import DEFAULT_RETRIES
from greenbone.scap.cpe_match.cli.processor import CPE_MATCH_TYPE_PLURAL
from greenbone.scap.errors import ScapError
from greenbone.scap.generic_cli.producer.nvd_api import NvdApiProducer


class CpeMatchNvdApiProducer(NvdApiProducer[CPEMatchString]):
    """
    Async context manager class for a producer querying
    CPE match strings from an NVD API.
    """

    _item_type_plural = CPE_MATCH_TYPE_PLURAL
    "Plural form of the type of items to use in log messages"

    _arg_defaults = NvdApiProducer._arg_defaults
    "Default values for optional arguments."

    @classmethod
    def from_args(
        cls,
        args: Namespace,
        console: Console,
        error_console: Console,
        progress: Progress,
    ) -> "CpeMatchNvdApiProducer":
        """
        Create a new `CPEMatchNvdApiProducer` with parameters from
         the given command line args gathered by an `ArgumentParser`.

        Args:
            args: Command line arguments to use
            console: Console for standard output.
            error_console: Console for error output.
            progress: Progress bar renderer to be updated by the producer.

        Returns:
            The new `CPEMatchNvdApiProducer`.
        """
        request_filter_opts = {}

        since = NvdApiProducer.since_from_args(args, error_console)
        if since is not None:
            request_filter_opts["last_modified_start_date"] = since

        return CpeMatchNvdApiProducer(
            console,
            error_console,
            progress,
            nvd_api_key=args.nvd_api_key,
            retry_attempts=args.retry_attempts,
            request_results=args.number,
            request_filter_opts=request_filter_opts,
            start_index=args.start,
            verbose=args.verbose or 0,
        )

    def __init__(
        self,
        console: Console,
        error_console: Console,
        progress: Progress,
        *,
        retry_attempts: int = DEFAULT_RETRIES,
        nvd_api_key: str | None = None,
        request_results: int | None = None,
        request_filter_opts: dict = {},
        start_index: int = 0,
        verbose: int | None = None,
    ):
        """
        Constructor for a CPE match string NVD API producer.

        Args:
            console: Console for standard output.
            error_console: Console for error output.
            progress: Progress bar renderer to be updated by the producer.
            retry_attempts: Number of retries for downloading items
            nvd_api_key: API key to use for the requests to allow faster requests
            request_results: Total number of results to request from the API
            request_filter_opts: Filter options to pass to the API requests
            start_index: index/offset of the first item to request
            verbose: Verbosity level of log messages.
        """
        self._nvd_api: CPEMatchApi

        super().__init__(
            console,
            error_console,
            progress,
            retry_attempts=retry_attempts,
            nvd_api_key=nvd_api_key,
            request_results=request_results,
            request_filter_opts=request_filter_opts,
            start_index=start_index,
            verbose=verbose,
        )

    def _create_nvd_api(self, nvd_api_key: str | None) -> CPEMatchApi:
        """
        Callback used by the constructor to create the NVD API object
        that can be queried for CPE match strings.

        Args:
            nvd_api_key: An optional API key to allow faster requests.

        Returns: The new `CPEMatchApi` object, which inherits from `NVDApi`.
        """
        return CPEMatchApi(
            token=nvd_api_key,
        )

    async def _create_nvd_results(self) -> NVDResults[CPEMatchString]:
        """
        Callback used during `fetch_initial_data` getting
        the `NVDResults` object the CPE match strings will be fetched from.

        Returns: The new `NVDResults` object.
        """
        if self._queue is None:
            raise ScapError("No queue has been assigned")

        return await self._nvd_api.cpe_matches(
            last_modified_start_date=self._request_filter_opts.get(
                "last_modified_start_date"
            ),
            last_modified_end_date=self._request_filter_opts.get(
                "last_modified_end_date"
            ),
            cve_id=self._request_filter_opts.get("cve_id"),
            match_string_search=self._request_filter_opts.get(
                "match_string_search"
            ),
            request_results=self._request_results,
            start_index=self._start_index,
            results_per_page=self._queue.chunk_size,
        )



# SPDX-FileCopyrightText: 2025 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from abc import abstractmethod
from argparse import ArgumentParser
from typing import (
    AsyncContextManager,
    AsyncIterator,
    Generic,
    Type,
    TypeVar,
)

from rich.console import Console
from rich.progress import Progress

from greenbone.scap.cli import (
    DEFAULT_POSTGRES_DATABASE_NAME,
    DEFAULT_POSTGRES_HOST,
    DEFAULT_POSTGRES_PORT,
    DEFAULT_VERBOSITY,
    CLIError,
)
from greenbone.scap.cpe_match.db.models import BaseDatabaseModel
from greenbone.scap.generic_cli.producer.base import BaseScapProducer

from ...db import PostgresDatabase
from ...timer import Timer

T = TypeVar("T")
"Generic type variable for the type of SCAP items handled"


class DatabaseProducer(BaseScapProducer, Generic[T]):
    """
    Abstract async context manager base class for a producer querying
    SCAP items from a PostgreSQL database.

    The type of items is to be defined by the generic type T in subclasses.
    """

    _item_type_plural = BaseScapProducer._item_type_plural
    "Plural form of the type of items to use in log messages"

    _arg_defaults = {
        "database_name": DEFAULT_POSTGRES_DATABASE_NAME,
        "database_host": DEFAULT_POSTGRES_HOST,
        "database_port": DEFAULT_POSTGRES_PORT,
        "database_schema": None,
        "verbose": DEFAULT_VERBOSITY,
    }
    "Default values for optional arguments."

    @classmethod
    def add_args_to_parser(
        cls: Type["DatabaseProducer"],
        parser: ArgumentParser,
    ):
        """
        Class method for adding database writer arguments to an
         `ArgumentParser`.

        Args:
            parser: The parser to add the arguments to.
        """
        db_group = parser.add_argument_group(
            title="Database", description="Database related settings"
        )

        db_group.add_argument(
            "--database-name",
            help=f"Name of the {cls._item_type_plural} database. "
            f"Uses environment variable DATABASE_NAME or "
            f"\"{cls._arg_defaults['database_name']}\" if not set.",
        )
        db_group.add_argument(
            "--database-host",
            help=f"Name of the {cls._item_type_plural} database host. "
            f"Uses environment variable DATABASE_HOST or "
            f"\"{cls._arg_defaults['database_host']}\" if not set.",
        )

        db_group.add_argument(
            "--database-port",
            help=f"Port for the {cls._item_type_plural} database. "
            f"Uses environment variable DATABASE_PORT or "
            f"{cls._arg_defaults['database_port']} if not set.",
            type=int,
        )
        db_group.add_argument(
            "--database-user",
            help=f"Name of the {cls._item_type_plural} database user. "
            f"Uses environment variable DATABASE_USER if not set.",
        )
        db_group.add_argument(
            "--database-password",
            help=f"Name of the {cls._item_type_plural} database password. "
            f"Uses environment variable DATABASE_PASSWORD if not set.",
        )
        db_group.add_argument(
            "--database-schema",
            help=f"Name of the {cls._item_type_plural} database schema. "
            f"Uses environment variable DATABASE_SCHEMA or "
            f"\"{cls._arg_defaults['database_schema']}\" if not set.",
        )
        db_group.add_argument(
            "--echo-sql",
            action="store_true",
            help="Print out all SQL queries.",
        )

    def __init__(
        self,
        console: Console,
        error_console: Console,
        progress: Progress,
        *,
        database_name: str | None,
        database_schema: str | None,
        database_host: str | None,
        database_port: int | None,
        database_user: str | None,
        database_password: str | None,
        echo_sql: bool = False,
        verbose: int = _arg_defaults["verbose"],
    ):
        """
        Constructor for a SCAP database write worker.

        If the `database_...` arguments are None or not given, corresponding
        environment variables will be tried next before finally using the
        defaults as a fallback.

        Args:
            console: Console for standard output.
            error_console: Console for error output.
            progress: Progress bar renderer to be updated by the producer.
            database_name: Name of the database to use.
            database_schema: Optional database schema to use.
            database_host: IP address or hostname of the database server to use.
            database_port: Port of the database server to use.
            database_user: Name of the database user to use.
            database_password: Password of the database user to use.
            echo_sql: Whether to print SQL statements.
            verbose: Verbosity level of log messages.
        """
        super().__init__(console, error_console, progress, verbose=verbose)

        database_name = (
            database_name
            or os.environ.get("DATABASE_NAME")
            or self._arg_defaults["database_name"]
        )
        database_schema = (
            database_schema
            or os.environ.get("DATABASE_SCHEMA")
            or self._arg_defaults["database_schema"]
        )
        database_host = (
            database_host
            or os.environ.get("DATABASE_HOST")
            or self._arg_defaults["database_host"]
        )
        try:
            port_str = os.environ.get("DATABASE_PORT")
            env_database_port = int(port_str) if port_str else None
        except TypeError:
            env_database_port = None
        database_port = (
            database_port
            or env_database_port
            or self._arg_defaults["database_port"]
        )
        database_user = database_user or os.environ.get("DATABASE_USER")
        database_password = database_password or os.environ.get(
            "DATABASE_PASSWORD"
        )

        if not database_user:
            raise CLIError(
                f"Missing user for {self._item_type_plural} database"
            )

        if not database_password:
            raise CLIError(
                f"Missing password for {self._item_type_plural} database"
            )

        self._database = PostgresDatabase(
            user=database_user,
            password=database_password,
            host=database_host,
            port=database_port,
            dbname=database_name,  # type: ignore
            schema=database_schema,
            echo=echo_sql,
        )
        if verbose:
            console.log(f"Using PostgreSQL database {database_name}")
        if verbose >= 2:
            console.log(
                f"Database host: {database_host}, port: {database_port}, "
                f"schema: {database_schema}, user: {database_user}"
            )

        self._manager = self._create_manager()

    @abstractmethod
    def _create_manager(self) -> AsyncContextManager:
        """
        Callback creating a new database manager for handling SCAP items.

        Returns: The new database manager.
        """
        pass

    @abstractmethod
    def _convert_db_model(self, db_model: BaseDatabaseModel) -> T:
        """
        Callback converting a SCAP database model to a Pontos model.

        Args:
            db_model: The database model convert

        Returns:
            The converted model object.
        """
        pass

    @abstractmethod
    async def _db_item_count(self) -> int:
        """
        Callback getting the total number of SCAP items in the database.

        Returns:
            The total number of items
        """
        pass

    @abstractmethod
    def _db_item_iter(self) -> AsyncIterator[BaseDatabaseModel]:
        """
        Callback getting an async iterator of database items to process.

        Returns:
            An async iterator over database items.
        """

    async def fetch_initial_data(self) -> int:
        """
        Ensures any remaining initializations of the data source are done,
         so it can be queried for chunks of SCAP items by `run_loop`.
        It must also return the expected total number of items that will be
         fetched.

        Returns:
            The expected total number of items.
        """
        count = await self._db_item_count()
        self._console.log(
            f"{count:,} {self._item_type_plural} available in database"
        )
        return count

    async def run_loop(self) -> None:
        """
        Run a loop fetching chunks of SCAP items and adding them to the queue.

        The method must also call `set_producer_finished` before returning to signal
        that no more chunks will be added to the queue.

        It should also create a task for the `progress` object and update it
        regularly.
        """
        task = self._progress.add_task(
            f"Querying {self._item_type_plural}",
            total=self._queue.total_items,
        )
        count = 0

        try:
            with Timer() as query_timer:
                chunk = []
                async for db_item in self._db_item_iter():
                    item: T = self._convert_db_model(db_item)
                    chunk.append(item)
                    if len(chunk) >= self._queue.chunk_size:
                        count += len(chunk)
                        await self._queue.put_chunk(chunk)
                        chunk = []
                    self._progress.update(task, completed=count)

                count += len(chunk)
                if len(chunk):
                    await self._queue.put_chunk(chunk)
                self._progress.update(task, completed=count)

            self._console.log(
                f"Queried {count:,} {self._item_type_plural} in "
                f"{query_timer.elapsed_time:0.4f} seconds."
            )

        finally:
            # signal worker that we are finished with querying the DB
            self._queue.set_producer_finished()

    async def __aenter__(self):
        await self._database.__aenter__()
        await self._manager.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        db_ret = await self._database.__aexit__(exc_type, exc_val, exc_tb)
        manager_ret = await self._manager.__aexit__(exc_type, exc_val, exc_tb)
        return db_ret or manager_ret



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest

from greenbone.scap.version import canonical_version, is_valid_version


class CanonicalVersionTestCase(unittest.TestCase):
    def test_invalid_version(self):
        self.assertIsNone(canonical_version("a.b"))
        self.assertIsNone(canonical_version("1.0def"))

    def test_canonical_version(self):
        self.assertEqual(canonical_version("1.2.3"), "1.2.3")
        self.assertEqual(canonical_version("1.2.3a1"), "1.2.3")
        self.assertEqual(canonical_version("1.2.3b1"), "1.2.3")
        self.assertEqual(canonical_version("1.2.3-rc1"), "1.2.3")
        self.assertEqual(canonical_version("1.2.3.dev1"), "1.2.3")


class IsValidVersionTestCase(unittest.TestCase):
    def test_invalid_version(self):
        self.assertFalse(is_valid_version(None))
        self.assertFalse(is_valid_version(""))
        self.assertFalse(is_valid_version("a.b"))
        self.assertFalse(is_valid_version("1.0def"))

    def test_valid_version(self):
        self.assertTrue(is_valid_version("1.2.3"))
        self.assertTrue(is_valid_version("1.2.3a1"))
        self.assertTrue(is_valid_version("1.2.3b1"))
        self.assertTrue(is_valid_version("1.2.3-rc1"))
        self.assertTrue(is_valid_version("1.2.3.dev1"))



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
from contextlib import redirect_stderr
from io import StringIO

from greenbone.scap.cpe.cli.find import parse_args


class ParseArgsTestCase(unittest.TestCase):
    def test_defaults(self):
        args = parse_args(["cpe"])

        self.assertFalse(args.echo_sql)
        self.assertEqual(args.verbose, 0)
        self.assertFalse(args.exact)
        self.assertIsNone(args.database_name)
        self.assertIsNone(args.database_host)
        self.assertIsNone(args.database_port)
        self.assertIsNone(args.database_user)
        self.assertIsNone(args.database_password)
        self.assertIsNone(args.database_schema)
        self.assertIsNone(args.version_start_including)
        self.assertIsNone(args.version_end_including)
        self.assertIsNone(args.version_start_excluding)
        self.assertIsNone(args.version_end_excluding)
        self.assertIsNone(args.limit)
        self.assertFalse(args.include_deprecated)

    def test_cpe(self):
        args = parse_args(["cpe"])

        self.assertEqual(args.cpe, "cpe")

    def test_database(self):
        args = parse_args(
            [
                "--database-name",
                "scap",
                "--database-host",
                "a-db-server",
                "--database-port",
                "123",
                "--database-user",
                "scap-user",
                "--database-password",
                "1234",
                "--database-schema",
                "scap-schema",
                "cpe",
            ]
        )

        self.assertEqual(args.database_name, "scap")
        self.assertEqual(args.database_host, "a-db-server")
        self.assertEqual(args.database_port, 123)
        self.assertEqual(args.database_user, "scap-user")
        self.assertEqual(args.database_password, "1234")
        self.assertEqual(args.database_schema, "scap-schema")

    def test_echo_sql(self):
        args = parse_args(["--echo-sql", "cpe"])

        self.assertTrue(args.echo_sql)

    def test_verbose(self):
        args = parse_args(["-v", "cpe"])

        self.assertTrue(args.verbose, 1)

        args = parse_args(["-vv", "cpe"])

        self.assertTrue(args.verbose, 2)

        args = parse_args(["-vvv", "cpe"])

        self.assertTrue(args.verbose, 3)

        args = parse_args(["--verbose", "cpe"])

        self.assertTrue(args.verbose, 1)

        args = parse_args(["--verbose", "--verbose", "cpe"])

        self.assertTrue(args.verbose, 2)

        args = parse_args(["--verbose", "--verbose", "--verbose", "cpe"])

        self.assertTrue(args.verbose, 3)

    def test_limit(self):
        args = parse_args(["--limit", "123", "cpe"])

        self.assertEqual(args.limit, 123)

        with self.assertRaises(SystemExit), redirect_stderr(StringIO()):
            parse_args(["--limit", "foo", "cpe"])

    def test_exact(self):
        args = parse_args(["--exact", "cpe"])

        self.assertTrue(args.exact)

    def test_include_deprecated(self):
        args = parse_args(["--include-deprecated", "cpe"])

        self.assertTrue(args.include_deprecated)

        args = parse_args(["--no-include-deprecated", "cpe"])

        self.assertFalse(args.include_deprecated)

    def test_version_start_including(self):
        args = parse_args(["--version-start-including", "1.2.3", "cpe"])

        self.assertEqual(args.version_start_including, "1.2.3")

    def test_version_end_including(self):
        args = parse_args(["--version-end-including", "1.2.3", "cpe"])

        self.assertEqual(args.version_end_including, "1.2.3")

    def test_version_start_excluding(self):
        args = parse_args(["--version-start-excluding", "1.2.3", "cpe"])

        self.assertEqual(args.version_start_excluding, "1.2.3")

    def test_version_end_excluding(self):
        args = parse_args(["--version-end-excluding", "1.2.3", "cpe"])

        self.assertEqual(args.version_end_excluding, "1.2.3")



# SPDX-FileCopyrightText: 2025 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
from typing import Sequence

from pontos.nvd.models.cpe_match_string import CPEMatchString
from rich.console import Console
from rich.progress import Progress

from greenbone.scap.generic_cli.worker.base import BaseScapWorker


class CpeMatchMockWorker(BaseScapWorker[CPEMatchString]):

    def __init__(
        self,
        console: Console,
        error_console: Console,
        progress: Progress,
        *,
        verbose: int | None = None,
    ):
        super().__init__(
            console=console,
            error_console=error_console,
            progress=progress,
            verbose=verbose,
        )
        self.context_entered: bool = False
        self.context_exited: bool = False
        self.items_received: list[CPEMatchString] = []
        self.item_count = 0
        self.chunk_count = 0

    async def _handle_chunk(self, chunk: Sequence[CPEMatchString]) -> None:
        self.items_received.extend(chunk)
        self.item_count += len(chunk)
        self.chunk_count += 1

    async def __aenter__(self):
        self.context_entered = True

    async def __aexit__(self, __exc_type, __exc_value, __traceback):
        self.context_exited = True



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
from time import sleep

from greenbone.scap.timer import Timer, TimerError


class TimerTestCase(unittest.TestCase):
    def test_context_manager(self):
        t = Timer()
        with t as t:
            sleep(0.1)

        self.assertIsNotNone(t.elapsed_time)

        with t as t:
            sleep(0.1)

        self.assertIsNotNone(t.elapsed_time)

    def test_timer(self):
        t = Timer()
        t.start()
        sleep(0.1)

        self.assertIsNotNone(t.stop())
        self.assertIsNotNone(t.elapsed_time)

        t.start()
        sleep(0.1)

        self.assertIsNotNone(t.stop())
        self.assertIsNotNone(t.elapsed_time)

    def test_start_twice(self):
        t = Timer()
        t.start()

        with self.assertRaises(TimerError):
            t.start()

    def test_stop_not_started(self):
        t = Timer()

        with self.assertRaises(TimerError):
            t.stop()



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
from datetime import datetime
from pathlib import Path

from greenbone.scap.cli import DEFAULT_RETRIES, DEFAULT_VERBOSITY
from greenbone.scap.cpe_match.cli.json_download import parse_args
from greenbone.scap.cpe_match.cli.processor import CPE_MATCH_DEFAULT_CHUNK_SIZE
from greenbone.scap.generic_cli.queue import DEFAULT_QUEUE_SIZE


class ParseArgsTestCase(unittest.TestCase):
    def test_defaults(self):
        args = parse_args([])

        self.assertIsNone(args.since)
        self.assertIsNone(args.since_from_file)
        self.assertIsNone(args.number)
        self.assertIsNone(args.start)
        self.assertEqual(DEFAULT_RETRIES, args.retry_attempts)
        self.assertIsNone(args.nvd_api_key)

        self.assertEqual(Path("."), args.storage_path)
        self.assertFalse(args.compress)

        self.assertEqual(CPE_MATCH_DEFAULT_CHUNK_SIZE, args.chunk_size)
        self.assertEqual(DEFAULT_QUEUE_SIZE, args.queue_size)
        self.assertEqual(DEFAULT_VERBOSITY, args.verbose)

    def test_since(self):
        args = parse_args(["--since", "2024-12-09"])
        self.assertEqual(datetime(2024, 12, 9), args.since)

    def test_since_from_file(self):
        args = parse_args(["--since-from-file", "/tmp/path"])
        self.assertEqual(Path("/tmp/path"), args.since_from_file)
